/*---------------------------------------------------------------------\
|                          ____ _   __ __ ___                          |
|                         |__  / \ / / . \ . \                         |
|                           / / \ V /|  _/  _/                         |
|                          / /__ | | | | | |                           |
|                         /_____||_| |_| |_|                           |
|                                                                      |
\---------------------------------------------------------------------*/
// Automatically generated wrapper for zypp/sat/SolvAttr.h
// Source Namespace: zypp -> Target Namespace: zyppng
#ifndef SOLVATTR_H_ZYPPNG_WRAPPER_H
#define SOLVATTR_H_ZYPPNG_WRAPPER_H

#include <zypp/sat/SolvAttr.h>

namespace zyppng
{

    namespace sat {
        using ::zypp::sat::SolvAttr;
    } // namespace sat

} // namespace zyppng
#endif // SOLVATTR_H_ZYPPNG_WRAPPER_H
