/*---------------------------------------------------------------------\
|                          ____ _   __ __ ___                          |
|                         |__  / \ / / . \ . \                         |
|                           / / \ V /|  _/  _/                         |
|                          / /__ | | | | | |                           |
|                         /_____||_| |_| |_|                           |
|                                                                      |
\---------------------------------------------------------------------*/
// Automatically generated wrapper for zypp/ResTraits.h
// Source Namespace: zypp -> Target Namespace: zyppng
#ifndef RESTRAITS_H_ZYPPNG_WRAPPER_H
#define RESTRAITS_H_ZYPPNG_WRAPPER_H

#include <zypp/ResTraits.h>

namespace zyppng
{

    namespace traits {
        using ::zypp::traits::isPseudoInstalled;
    } // namespace traits

    using ::zypp::Resolvable;
    using ::zypp::ResObject;
    using ::zypp::Package;
    using ::zypp::SrcPackage;
    using ::zypp::Pattern;
    using ::zypp::Product;
    using ::zypp::Patch;
    using ::zypp::Application;
    using ::zypp::PoolItem;
    using ::zypp::ResTraits;
    using ::zypp::resKind;
    using ::zypp::isKind;
} // namespace zyppng
#endif // RESTRAITS_H_ZYPPNG_WRAPPER_H
