/*---------------------------------------------------------------------\
|                          ____ _   __ __ ___                          |
|                         |__  / \ / / . \ . \                         |
|                           / / \ V /|  _/  _/                         |
|                          / /__ | | | | | |                           |
|                         /_____||_| |_| |_|                           |
|                                                                      |
\---------------------------------------------------------------------*/
// Automatically generated wrapper for zypp/RelCompare.h
// Source Namespace: zypp -> Target Namespace: zyppng
#ifndef RELCOMPARE_H_ZYPPNG_WRAPPER_H
#define RELCOMPARE_H_ZYPPNG_WRAPPER_H

#include <zypp/RelCompare.h>

namespace zyppng
{
    using ::zypp::Compare;
    using ::zypp::compareByRel;
    using ::zypp::CompareBy;
    using ::zypp::CompareByEQ;
    using ::zypp::CompareByNE;
    using ::zypp::CompareByLT;
    using ::zypp::CompareByLE;
    using ::zypp::CompareByGT;
    using ::zypp::CompareByGE;
    using ::zypp::CompareByANY;
    using ::zypp::CompareByNONE;
} // namespace zyppng
#endif // RELCOMPARE_H_ZYPPNG_WRAPPER_H
